"""
Simulate disk addition flow with high concurency, trying to reproduce
https://bugzilla.redhat.com/show_bug.cgi?id=1014942
"""

import threading
import subprocess
import logging
import optparse

class options:

    # Test setup

    iterations = 1
    concurrency = 1
    verbose = False
    debug = False
    use_filters = True

    # Machine setup - you must update this for your machine

    vg_name = 'e232d362-d3a9-4a03-87bf-7c9e06c2af31'

    devices = [
        '/dev/mapper/3600144f09dbd05000000522dde040011',
        '/dev/mapper/3600144f09dbd05000000522dde070013',
    ]

    # Common - configuration used by vdsm

    config = '''
devices {
 preferred_names=["^/dev/mapper/"]
 ignore_suspended_devices=1
 write_cache_state=0
 disable_after_error_count=3
 %(filter)s
}
global {
 locking_type=1
 prioritise_write_locks=1
 wait_for_locks=1
}
backup {
 retain_min=50
 retain_days=0
}
'''.replace('\n', ' ')

parser = optparse.OptionParser()
parser.add_option('-i', '--iterations', type='int')
parser.add_option('-c', '--concurrency', type='int')
parser.add_option('-v', '--verbose', action='store_true')
parser.add_option('-d', '--debug', action='store_true')
parser.add_option('-n', '--no-filters', dest='use_filters', action='store_false')

lock = threading.Lock()
running = True

class LVMError(Exception):
    pass

def is_running():
    global running
    with lock:
        return running

def stop():
    global running
    with lock:
        running = False

def make_filter(devices):
    if options.use_filters and devices:
        return 'filter=["a|%s|", "r|.*|"]' % '|'.join(devices)
    else:
        return ''

def quote(arg):
    if len(arg.split()) > 1:
        arg = arg.replace("'", "\\'")
        arg = "'" + arg + "'"
    return arg

def lvm(name, *args, **kw):
    cfg = options.config % {'filter': make_filter(kw.get('devices'))}
    cmd = ["/sbin/lvm", name, "--config", cfg]
    if options.verbose:
        cmd.append('-vvvv')
    cmd.extend(args)
    command = ' '.join((quote(a) for a in cmd))
    logging.debug('running: %s', command)
    proc = subprocess.Popen(cmd, close_fds=True, bufsize=4096,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = proc.communicate()
    if proc.returncode != 0:
        msg = name + ' error\n'
        msg += 'command\n  ' + command + '\n'
        msg += 'stdout\n' + out + '\n'
        msg += 'stderr\n' + err + '\n'
        raise LVMError(msg)

def create_volume(lv_name):
    lv_path = options.vg_name + '/' + lv_name

    logging.info('creating %r', lv_path)

    lvm("lvcreate",
        "--autobackup", "n",
        "--contiguous", "n",
        "--size", "1024m",
        "--name", lv_name,
        options.vg_name,
        devices=options.devices
    )

    lvm("lvchange", "--available", "n", lv_path, devices=options.devices)

    lvm("lvs", options.vg_name, devices=options.devices)

    logging.info('removing %r', lv_path)

    lvm("lvremove", "--autobackup", "n", lv_path, devices=options.devices)

def worker(name):
    count = 0
    try:
        logging.info('worker %r: starting', name)
        while is_running() and count < options.iterations:
            count += 1
            lv_name = '%s-%02d' % (name, count)
            create_volume(lv_name)
        logging.info('worker %r: done', name)
    except Exception as e:
        logging.error('worker %r: failed: %s', name, e)
        stop()


_, args = parser.parse_args(values=options)

log_level = logging.DEBUG if options.debug else logging.INFO
logging.basicConfig(level=log_level,
                    format='%(asctime)s %(levelname)s %(message)s')

workers = []

logging.info('starting %d workers', options.concurrency)

try:
    for i in range(options.concurrency):
        name = 'test-%02d' % i
        t = threading.Thread(target=worker, args=(name,))
        t.start()
        workers.append(t)
except KeyboardInterrupt:
    logging.info('interrupted')
    stop()

for w in workers:
    w.join()

logging.info('done')
